from .objects import COCO_Info, COCO_License, COCO_Image, \
    COCO_Annotation, COCO_Category
from .handlers import COCO_License_Handler, COCO_Image_Handler, \
    COCO_Annotation_Handler, COCO_Category_Handler
from .dataset import COCO_Dataset